# -*- coding: utf-8 -*-
"""
Created on Fri Sep 10 13:07:49 2021

@author: NeuroBrave
"""

import json
import random
import string


class UtilService:
    
    def __init__(self):
        pass
      
    def load_config_file(config_path): 
        with open(config_path) as f:
          config = json.load(f)
          return config
      
        
    def generateId(id_length):
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(id_length))



# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""


from pathlib import Path
import os
from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService

import socket
import json

#the following imports needed only to operate NeuroSpeed.IO cloud recorder:
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.macros import macros
import time


global user_auth


UDP_IP="127.0.0.1"
UDP_PORT = 5005
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

def EEG_processing_handler(payload):
    '''
    this function extracts data from payload each time a packet is received - typically 1 packet every second.
    Parameters
    ----------
    payload : dictionary with values of raw sensor data, processed information and processing parameters.

    Returns
    -------
    None.

    payload structure: see file "eeg_payload.json"


    example  payload["sensor_info"] structure: 
    	{
		'id': 6461,
		'device_type': 'eeg',
		'sensor_id': '9LUMVF'
		'stream_id': 'EEG_9LUMVF_0',
		'is_attached': True,
		'sensor_version': 1, 
		'stream_state': 'enabled',
		'channels_count': 4,
		'created_at': '2022-06-01T14:13:49.926Z',
		'channel_map': ['TP9', 'AF7', 'AF8', 'TP10', 'timestamp'], 
		'buffer_length': 1,
		'manufacturer_id': 'muse',
		'sampling_frequency': 256, 
		'buffer_length_samples': 256,
		'buffer_length_seconds': 1
		}
 
    '''
   
    # to get raw sensor data:
    # raw_data = payload["data"]["sample"]        #returns list with 256 entries, each entry is list of 5 values: 4 channel voltages and a timestamp
    
    #to get cleaned raw data: 
    # cleaned_data = payload["output"]["raw_data"]["clean"]   #returns list with 256 entries, each entry is list of 4 values: 4 channel voltages 

    #to get neuro-marker data:         
    # marker_name = 'engagement_upgraded_index_h-b'
    # marker_data = payload["output"][marker_name]   #returns dictionary {'hard_decision': <int>, 'soft_decision_v1': <float>, 'soft_decision_v2': <float>},

    # #to get channel health indicators: 
    # bad_channel = payload["output"]['bad_channel'] #returns list with 4 entries, each entry is an integer: 0 if channel is good, 1 if channel is bad.

    print("received NeuroSpeed data packet")

    output_packet = {}
    output_packet["clean_data"] = payload["raw_data"]["clean"] 
    output_packet["brainwave_power"] = payload["output"]["brainwave_power"]
    output_packet["engagement_upgraded_index_h-b"] = payload["output"]["engagement_upgraded_index_h-b"]
    output_packet["focus_attention_upgraded_index_h-b"] = payload["output"]["focus_attention_upgraded_index_h-b"]
    output_packet["engagement_level_digital_media_upgraded_h-b"] = payload["output"]["engagement_level_digital_media_upgraded_h-b"]
    output_packet["bad_channel"] = payload["output"]["bad_channel"]
    # print(payload) 

    to_send = str(output_packet).encode('utf-8')
    if len(to_send) > 65000:
        print("UDP packet size exceded, not sending")
    else:
        sock.sendto(to_send, (UDP_IP, UDP_PORT))
    
    

generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler],
    'gamepad': [],
    'user_data': [],
    'ppg': [],
    'keyboard':[]
}



def user_data_external_handler(payload):
    username = user_auth.get_username()
    print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, payload["hia_id"], payload["stream_id"], payload["device_type"]))
    
    # execute each function handler for relevant device_type
    device_type = payload["device_type"].lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)
    if payload["type"] =="disconnect": 
        print(f"client {payload['hia_id']} for user {payload['username']} disconnect from sensor detected!")
    if payload["type"] =="connect": 
        print(f"client {payload['hia_id']} for user {payload['username']} disconnect from sensor detected!")
        



# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    
    
    #the following code is used to connect to user-level API to subscribe to data stream:
    global user_auth
    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config\\","hia_config.json"))
    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    
    # the following 2 lines of code are optional: subscribe to connect/disconnect events to manage related events:
    userRoom.set_device_events_external_handler(user_device_event_external_handler)
    userRoom.connect()


    # this optional chunk of code is used to start the data recorder on neurospeed.io
    with open('config/dashboard_credentials.json') as f:
         customer_config = json.load(f)             
    # Authenticate as customer on neurospeed.io:
    print(f"using customer config {customer_config}")
    customer_auth = Auth_AS_Customer_handler(customer_config)
    login_success = customer_auth.login()
    
    if not login_success:
        print("Failed to log in as Customer")
    else:
        print("logged in as Customer, initiating recorder")
        recording_name  = "my_recording"
        recorder_id = macros.start_recording(customer_auth, config_user1, recorder_name = recording_name)
    
    
    
        time.sleep(60)
        
        
        
        
        # stop the data recording on the cloud. Stopping the recording does not automatically download the data, another command does it.
        macros.stop_recording(recorder_id)
     
        # command the cloud engine to perform the data export and download the datafile:  
            
        # "exporter" finalizes the recording, converts it to human-readable .csv file and makes it available for download:
        # exporter name can be any string. exporter can export certain range of the recording, by selecting start and end timestamps. 
        # to export the entire recording, use: "start_timestamp_mode": "start", "end_timestamp_mode": "end"
     
        exporter_config =  {"custom_name": recording_name + "_exported",    
        "start_timestamp_mode": "start", 
        "end_timestamp_mode": "end"}
        
        save_folder =  os.getcwd()        
        exporter_filename = macros.download_data(customer_auth, recorder_id, exporter_config, config_user1, save_folder)
        
        
    
   

if __name__ == '__main__':    
    main()  




from neurospeed.utils.helper_service import UtilService as utils
from pathlib import Path
import os

class ApiConfig:

    def __init__(self):
        self._contex = "ApiConfig - "
        
        '''
        precedence of config path options: 
        
        if C:/neurobrave/hia_config.json exists, it's used
        otherwise, if /config/api_config.json exists under the runtime folder, its' used
        it none of these exist, defauilt that comes with the neurospeed library is used.
            
        '''
        # if os.path.isfile("C://neurobrave//HIA_config.json"):
        #     self._config_path = "C://neurobrave//HIA_config.json"
        #     print("using HIA config for API URL")
        # elif os.path.isfile(os.path.join(os.getcwd(),"config", "api_config.json")):
        #     self._config_path = os.path.join(os.getcwd(),"config", "api_config.json")
        #     print("using program config for API URL")
        # else:     
        #     self._config_path =  os.path.join(str(Path(__file__).parent.parent) ,"config","api_config.json")

        if os.path.isfile(os.path.join(os.getcwd(),"config", "api_config.json")):
            self._config_path = os.path.join(os.getcwd(),"config", "api_config.json")
            print("using program config for API URL")
        else:     
            self._config_path =  os.path.join(str(Path(__file__).parent.parent) ,"config","api_config.json")


        self._api_config = utils.load_config_file(self._config_path)
        self._api_http_url =  self._api_config["api_address_prod"]  if self._api_config["is_prod"] == "True"  else  self._api_config["api_address_local"] 
        self._pipeline_url =  self._api_config["pipeline_address_prod"]  if self._api_config["is_prod"] == "True"  else  self._api_config["pipeline_address_local"] 
        print("api config path: " + self._config_path)
        print("{} NeuroSpeed API HTTP URL: {} ".format(self._contex, self._api_http_url))
        print("{} NeuroSpeed API SOCKET URL: {} ".format(self._contex, self._pipeline_url))
        
    def get_http_url(self):
        return self._api_http_url
    
    def get_socket_url(self):
        return self._pipeline_url
        



# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""

import time
from pathlib import Path
import os
from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
import socket
import json


global user_auth
bad_channel_statistics={}
def EEG_processing_handler1(payload):
    '''
    this function extracts data from payload each time a packet is received - typically 1 packet every second.
    Parameters
    ----------
    payload : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    example payload["output"] structure:
        {'heart_rate': {'estimated_heart_rate': 59.53488372093023}, 'ICA': {'artifact_detected': 0, 'artifact_power': -1}, 'bad_channel': [0, 0, 0, 0], 'engagement_upgraded_index_h-b': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_manf_index_h-p': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_upgraded_index_h-p': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_manf_index_h-b': {'hard_decision': 0, 'soft_decision_v1': 0, 'soft_decision_v2': 0}, 'engagement_upgraded_index_h-p': {'hard_decision': 0, 'soft_decision_v1': 0.012534856796264648, 'soft_decision_v2': 0.025069713592529297}, 'focus_attention_upgraded_index_h-b': {'hard_decision': 0, 'soft_decision_v1': 0, 'soft_decision_v2': 0}}

    example  payload["sensor_info"] structure: 
    
    '''
     

    # EEG_sensor_information = {}
    # EEG_sensor_information["EEG_channel_names"] = payload["sensor_info"]["channel_map"]
    # EEG_sensor_information["EEG_channel_num"] = len(payload["sensor_info"]["channel_map"])    
    # EEG_sensor_information["sampling_frequency"] = payload["sensor_info"]["sampling_frequency"]
    # print(EEG_sensor_information)

    print("got EEG packet")
    # print(payload["output"])
    
    # print(payload["output"]['bad_channel'])
      
def ppg_handler_func(payload):
    # print(payload["output"]["heart_rate_ppg"])
    pass
    

def EEG_processing_handler2(payload):
    # your implementation
    # print('EEG_processing_handler2')
    print(payload["output"].keys())
    pass
    
def customer_gamepad_msg_handler_function(payload):
    # your implementation
    print('customer_gamepad_msg_handler_function')
    pass

def customer_user_data_msg_handler_function(payload):
    # your implementation
    print('customer_user_data_msg_handler_function')
    pass


def smartwatch_handler_func(payload):
    print("received smartwatch data: {payload.keys()}")
    
    
generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler1, EEG_processing_handler2],
    'gamepad': [customer_gamepad_msg_handler_function],
    'user_data': [customer_user_data_msg_handler_function],
    'ppg': [ppg_handler_func],
    'smartwatch': [smartwatch_handler_func]
}



def user_data_external_handler(payload):
    username = user_auth.get_username()
    stream_id = payload["stream_id"]
    device_type = payload["device_type"]
    hia_id = payload["hia_id"]
    sensor_info = payload["sensor_info"]
    print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, hia_id, stream_id, device_type))
    
    # execute each function handler for relevant device_type
    device_type = payload["device_type"].lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)
    pass;



# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    global user_auth


    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config\\","hia_config.json"))

    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    userRoom.set_device_events_external_handler(user_device_event_external_handler)

    userRoom.connect()

   

if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 14 13:42:26 2021

@author: NeuroBrave
"""


import time
from pathlib import Path
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.api_http_handlers.exporter_http_handler import UserRoom_Recorder_Exporter_Handler
from neurospeed.utils.helper_service import UtilService

import os
global customer_auth
global exporter_handler
global username


# once recorder has at least 1 packet recorded, you can create exporter to export data
# exporter which belongs to some recorder, must finish first before another exporter can be created for this recorder
def run_exporter_flow(recorder_id, exporter_config):
    global exporter_handler
    
    # create exporter, if recorder has no data, request will fail.
    exporter = exporter_handler.create_exporter(recorder_id, exporter_config)
    exporter_id = str(exporter["id"])
    
    # once exporter created, query it's status until it "exported".
    exporter_exported = False
    max_query_attemps = 20
    query_attemps = 1
    while (exporter_exported != True & (query_attemps < max_query_attemps) ):
        exporter = exporter_handler.get_exporter(exporter_id)   
        print("exporter status: " + exporter["status"])
        if (exporter["status"] == "exported"):
            print('exporter finished exporting')
            exporter_exported = True
        else:
            print('exporter not yet exported the files')
            query_attemps = query_attemps + 1
            time.sleep(10)
             
    if (exporter_exported != True):
        print('exporter not yet exported the file after  ' + str(query_attemps)  + ' query attempts')
        exporter_handler.delete_exporter(exporter_id)
        return
   
    # once exporter exported the file, you can ask for an URL to download the exported file. 
    #the url available for only short time before it deprecates
    url_response = exporter_handler.get_exported_url(exporter_id)
    print(url_response)
    
    time.sleep(2)
   #delete_exporter(exporter_id) # uncomment this to delete exporter
    time.sleep(1)
    
    # list all exporters under username
    page_number = 1 
    page_size = 5 # max 100 rows per page
    exporters = exporter_handler.list_exporters(username, page_number, page_size)
    print(exporters)
    
    
def main():
    global customer_auth
    global recorder_handler
    global exporter_handler
    global username

     # load Customer configuration
    customer_config_path = os.path.join(str(Path().absolute()) ,"config","customer_config.json")
    customer_config = UtilService.load_config_file(customer_config_path)
    
    # Authenticate as customer
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()
    
    exporter_handler = UserRoom_Recorder_Exporter_Handler(customer_auth)
    
    
    recorder_id  = 12 # !!! set here the recorder_id 
    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config","hia_config1.json"))
    username = config_user1["username"] # or set here another custom username
  
    # exporter configuration
    exporter_config =  {
        # name your exporter up to 128 chars
        "custom_name": "export me gently", 
        
         # start_timestamp_mode. values options: 
         # 1) "start" - from the first recorded timestamp.  
         # 2) "custom_value" - custom starting timestamp in seconds(!) . 
         # for example : "1629900352" or for better Accuracy "1629900352.11..." 
         # but not 162990035211" which is timestamp represented in miliseconds instead of seconds!
        "start_timestamp_mode": "start", 
        
         #  custom_start_timestamp required if "start_timestamp_mode" is "custom_value"
        # "custom_start_timestamp": 1629900352.11111, 
        
        # end_timestamp_mode. values options:
        # 1) "end" - until the last recorded timestamp  
        # 2) "custom_value" - custom end timestamp in seconds(!). same logic as "start_timestamp_mode"
        # 3) interval_minutes - interval in minutes. set
        "end_timestamp_mode": "end", 
        
        # custom_timestamp_interval required if "end_timestamp_mode" is "interval_minutes", 
        #"custom_timestamp_interval": 3 ,
        
        # custom_end_timestamp required if "end_timestamp_mode" is "custom_value", 
       # "custom_end_timestamp": "1629900353.11111", 
       
       # to exporting specific streams, add their stream_id into the array.  
       # to exporting all recorded streams skip this configuration
       # "stream_ids": ["stream_id1", "stream_id2"]
    }
    run_exporter_flow(recorder_id, exporter_config) 


if __name__ == '__main__':    
    main()  



'''

this script sends data from EEG raw recordings in CSV format to neurospeed, and downloads the processed output files
the scripts expects a /config subfolder with following files:
    dashboard_credentials.json - for customer-level auth (to operate the recorder-exporter function)
        inside:
            {
            "email": "",
            "password": ""
            }


    HIA_config.json - for user-level auth (to send the EEG data)
        inside:
            {
                "account_id": "neurobrave_Nt57tr0rSG",
                "username": "oleg@neurobrave",
                "user_password": "oleg123321",
            }
    optional: api_config.json (if exists, the neurospeed server address will be used from this file. useful for running on non-default neurospeed servers.)
        inside:

            {
            "is_prod": "True",
            "api_address_prod": "https://api-test.neurospeed.io",
            "api_address_local":  "http://localhost:3000",
        	"pipeline_address_prod": "wss://api-test.neurospeed.io",
        	"pipeline_address_local": "ws://localhost:3000",
            }

    the scripts expects a "raw_data_folder" subfolder in it's root location (current working directory)
    in that "raw_data_folder" subfolder expected to be separate folder for each subject's experiment.
    the list "tasks" defines in which order to send the datafiles to neurospeed, accoring to partial match of data file names
    to keyword in the list
    the script saves to the "raw_data_folder" an exported datafle downloaded from NeuroSpeed; with labels according to the keywords.

    the data files expected to be in formats:
        CSV comma separated file

        data columns:
        timestamps, EEG channel 0,..EEG channel N

        timestamps are discarded when sending data


'''
import json
import numpy as np
import os
import sys
import time
import logging
import requests
from datetime import datetime
import pandas as pd

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
# from neurospeed.api_socket_handlers.downlink_room_as_user_handler import DownlinkRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
from neurospeed.hia_user_data.neurobrave_sensor_interface import HIA_Client

from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
# from neurospeed.api_http_handlers.recorder_http_handler import UserRoom_Recorder_Handler
# from neurospeed.api_http_handlers.exporter_http_handler import UserRoom_Recorder_Exporter_Handler


from neurospeed.macros import macros



def generate_sensor_info(config, hia_config):
    '''
    this functiuon generates sensor_info dictionary that is nesessary ot initialize the HIA client.
    the sensor info is tailored for single EEG according to config
    only LSL interface supported currently

    input:
        hia_config dictionary
    '''
    # generate SOME sensors stream ids
    sensor_info = dict()

    if "EEG" in config["DATA_SOURCE"]:
        user_data_stream_id = "EEG" + '_' + UtilService.generateId(6)
        sensor_info[user_data_stream_id] = {
            "device_type": "EEG",
            "channel_count": len(hia_config["EEG_channel_enable"]),
            "sampling_frequency": hia_config["EEG_sample_rate"],
            "buffer_length": hia_config["EEG_buffer_length_seconds"],
            "manufacturer_id": "NeuroBrave",
            "sensor_id": user_data_stream_id,
            "stream_id": user_data_stream_id,
            "stream_state": "enabled",
            "channel_map": hia_config["channel_map"]
        }

    if "PPG" in config["DATA_SOURCE"]:
        user_data_stream_id = "PPG" + '_' + UtilService.generateId(6)
        sensor_info[user_data_stream_id] = {
            "device_type": "PPG",
            "channel_count": len(hia_config["PPG_channel_map"]),
            "sampling_frequency": hia_config["PPG_sample_rate"],
            "buffer_length": hia_config["PPG_buffer_length_seconds"],
            "manufacturer_id": "NeuroBrave",
            "sensor_id": user_data_stream_id,
            "stream_id": user_data_stream_id,
            "stream_state": "enabled",
            "channel_map": hia_config["PPG_channel_map"]
        }

    if "SMARTWATCH" in config["DATA_SOURCE"]:
        user_data_stream_id = "PPG" + '_' + UtilService.generateId(6)
        sensor_info[user_data_stream_id] = {
            "device_type": "SMARTWATCH",
            "channel_count": len(hia_config["PPG_channel_map"]),
            "sampling_frequency": hia_config["PPG_sample_rate"],
            "buffer_length": hia_config["PPG_buffer_length_seconds"],
            "manufacturer_id": "NeuroBrave",
            "sensor_id": user_data_stream_id,
            "stream_id": user_data_stream_id,
            "stream_state": "enabled",
            "channel_map": hia_config["PPG_channel_map"]
        }


    if "GSR" in config["DATA_SOURCE"]:
        for idx, port in enumerate(config["OPENBCI_COM_PORT"]):
            user_data_stream_id = "GSR" + '_' + UtilService.generateId(6)
            sensor_info[user_data_stream_id] = {"device_type": "GSR",
                                                "sensor_id": user_data_stream_id,
                                                "stream_id": user_data_stream_id,
                                                "sampling_frequency": 250,
                                                "buffer_length": 1.0,
                                                "manufacturer_id": "openBCI",
                                                "channel_count": len(config["GSR_channel_enable"][idx]),
                                                # "Downstream ID":downstream_id,
                                                "channel_map": config["GSR_channel_map"][idx],
                                                "stream_state": "enabled"}

    if "EDA" in config["DATA_SOURCE"]:
        user_data_stream_id = "GSR_ADIEDA"

        sensor_info[user_data_stream_id] = {"device_type": "GSR",
                                            "sensor_id": user_data_stream_id,
                                            "stream_id": user_data_stream_id,
                                            "sampling_frequency": config["EDA_sampling_frequency"],
                                            "buffer_length": 1.0,
                                            "manufacturer_id": "ADI",
                                            "channel_count": 6,
                                            # "Downstream ID":downstream_id,
                                            "channel_map": ['real', 'img', 'adm_real', 'adm_img', 'adm_mag',
                                                            'adm_phase', "timestamp"],
                                            "stream_state": "enabled"}
    return sensor_info


def disconnect_external_handler(hia_client):
    # logging.debug("hia:["+ hia_client.get_hia_id()+"]"+ "user:["+ hia_client.get_username()+"] disconnected ")
    pass


# would be called for each hia after successfuly connection
def connection_external_handler(hia_client):
    pass
    # logging.debug('connected as '+ hia_client.get_username()+ " with hia: " +hia_client.get_hia_id()+ ". sending data..")


def init_HIA_client(config, hia_config):
    '''
    this functiuon  initialize the HIA client and connect the cloud streaming (upload raw data to neurospeed)
    the sensor info is tailored for single EEG according to config
    only LSL interface supported currently

    input:
        config - dictionary with config for this epxeriment program
        hia_config dictionary

    output
        hia_user client object instance
        sensor info dictionary

    '''
    user1_auth = Auth_AS_User_Handler(hia_config)
    user1_auth.login()

    hia_sensor_info_user1 = generate_sensor_info(config, hia_config)

    logging.debug('Generated sensor info: {}'.format(hia_sensor_info_user1))
    hia_user1 = HIA_Client(user1_auth, hia_sensor_info_user1)
    hia_user1.set_socket_connection_external_handler(connection_external_handler)
    hia_user1.set_socket_disconnect_external_handler(disconnect_external_handler)
    logging.debug("HIA_ID: " + str(user1_auth.get_hia_id()))
    hia_user1._device_type = "PPG"
    hia_user1.connect()
    return hia_user1, hia_sensor_info_user1



def send_EEG_to_websocket_from_files_in_datafolder(datafolder, hia_client, hia_sensor_info_user1, hia_config,
                                                   datatypes=["EEG"]):
    df_exporter = None
    for item in os.listdir(datafolder):
        if "_exported" in item and ".csv" in item:
            path = os.path.join(datafolder, item)
            df_exporter = pd.read_csv(path, delimiter=",", usecols=[
                'timestamp', 'exp.0', 'input.TP9', 'input.AF7', 'input.AF8', 'input.TP10', 'stream_id'
            ])
            break
    if df_exporter is None:
        raise ValueError("no Exporter file found!!!")
        # stream = df_exporter["stream_id"].unique()[0]
    print(df_exporter['stream_id'].dropna().unique())
    src_stream = [stream_id for stream_id in df_exporter['stream_id'].dropna().unique() if 'EEG_' in stream_id][-1]
    stream = list(filter(lambda x: x.startswith(datatypes[-1]), list(hia_sensor_info_user1.keys())))[0]

    df_eda = df_exporter.query(f"stream_id == @src_stream")
    eeg_labels_array = df_eda["exp.0"].to_numpy()
    data_array = df_eda.drop(columns=['timestamp', 'exp.0', 'stream_id']).to_numpy()

    stream_id = stream
    old_key = list(hia_sensor_info_user1.keys())[-1]
    hia_sensor_info_user1[stream] = hia_sensor_info_user1.pop(old_key)
    window = 256
    read_index = 0
    current_label = None
    print(f"length of data: {len(data_array)}")
    print(f"shape of the array is {data_array.shape}")
    while read_index + window < len(data_array):
        data = data_array[read_index:read_index + window].tolist()
        read_index += window
        hia_client.send_data_direct(data, stream_id, send_without_timestamp=True, device_type="EEG")
        print("tranmsmitting data packet")
        time.sleep(0.1)
        if data_array[read_index] != current_label:
            hia_client.set_label(eeg_labels_array[read_index])
            current_label = eeg_labels_array[read_index]
            print(f"setting label to {current_label}")


if __name__ == '__main__':




    config = {}
    config["DATA_SOURCE"] = ["EEG"]
    config["patientname"] = "olegtest"


    #put here a path to folder that has a CSV file in neurospeed format, that has "exported" in it's name: 
    datafolder = "H:\\DATASETS\\Simon-Waldo-Prime-Fix-Relax\\Experiment Recording\\Test\\data"

    hia_client = None
    with open('config/hia_config.json') as f:
        hia_config = json.load(f)

    with open('config/dashboard_credentials.json') as f:
        customer_config = json.load(f)
    # Authenticate as customer on neurospeed.io:
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()


    recorder_id = macros.start_recording(customer_auth, hia_config, recorder_name=config["patientname"])

    max_hia_reconnect_retries = 3
    for reconnect_idx in range(max_hia_reconnect_retries):
        hia_client, hia_sensor_info = init_HIA_client(config, hia_config)
        if hia_client.is_connected():
            break
        else:
            logging.debug("HIA connection retrying...")
            time.sleep(3)

    if not hia_client.is_connected():
        print("FAILED connecting to cloud streaming!\n")
        sys.exit()


   
    if 'EEG' in config["DATA_SOURCE"]:
        send_EEG_to_websocket_from_files_in_datafolder(datafolder, hia_client, hia_sensor_info,
                                                       hia_config,
                                                       config["DATA_SOURCE"])

    hia_client.disconnect()
    print("hia status: " + str(hia_client.is_connected()))
    macros.stop_recording(recorder_id)

    exporter_config = {
        "custom_name": config["patientname"] + "_exported",
        "start_timestamp_mode": "start",
        "end_timestamp_mode": "end"
    }
    save_folder = os.path.join(os.getcwd(), datafolder)
    exporter_filename = macros.download_data(customer_auth, recorder_id, exporter_config, hia_config,
                                             save_folder)



'''

this script starts cloud-based recorder on neursospeed,
sets experiment labels to create labeled data, 
sends some random raw data from 
and downloads the processed output files

the scripts expects a /config subfolder with following files:
    dashboard_credentials.json - for customer-level auth (to operate the recorder-exporter function)
        inside:
            {
            "email": "",
            "password": ""
            }
    
            
    HIA_config.json - for user-level auth (to send the data)
        inside: 
            {
                "account_id": "",
                "username": "",
                "user_password": "",
                "HIA_ID": "XYZ123",
                "Verbose_socket_log": "False"
            }
            
    optional: api_config.json (if exists, the neurospeed server address will be used from this file. useful for running on non-default neurospeed servers.)
    inside: 
            
           {
           "is_prod": "True", 
           "api_address_prod": "https://api.neurospeed.io",
           "api_address_local":  "http://localhost:3000",
           	"pipeline_address_prod": "wss://api.neurospeed.io",
           	"pipeline_address_local": "ws://localhost:3000",
           }
                 
'''
import json
import os
import sys
import time
import logging
from datetime import datetime

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
from neurospeed.hia_user_data.neurobrave_sensor_interface import HIA_Client
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.macros import macros


def generate_sensor_info():
    '''
    this function generates sensor_info dictionary that is nesessary ot initialize the HIA client.
    input:
        none
    '''
    # generate  sensors stream ids
    
    sensor_info = dict()
    user_data_stream_id =  "user_data" + '_' + UtilService.generateId(6)
    sensor_info[user_data_stream_id] =  {
    "device_type": "user_data",
    "channel_count": 2,
    "sampling_frequency": 10,
    "buffer_length" : 1.0,  #data buffer length in seconds. 
    "manufacturer_id": "myCompanyName",
    "sensor_id" : user_data_stream_id,
    "stream_id": user_data_stream_id,
    "stream_state" : "enabled",
     "channel_map" : ["chanllel_0", "channel_1"]
    } 
    
    user_data_stream_id =  "some_custom_stream_id"
    sensor_info[user_data_stream_id] =  {
    "device_type": "user_data",
    "channel_count": 2,
    "sampling_frequency": 10,
    "buffer_length" : 1.0,  #data buffer length in seconds. 
    "manufacturer_id": "myCompanyName",
    "sensor_id" : user_data_stream_id,
    "stream_id": user_data_stream_id,
    "stream_state" : "enabled",
     "channel_map" : ["chanllel_0", "channel_1"]
     }

    return sensor_info


def disconnect_external_handler(hia_client):
    # this function is automatically executed when this client disconnects from NeuroSpeed cloud
    pass
    

# would be called for each hia after successfuly connection  
def connection_external_handler(hia_client):
    # this function is automatically executed when this client connects to NeuroSpeed cloud
    pass

def init_HIA_client(hia_config):
    
    '''
    this function  initializes the HIA client and connect the cloud streaming (upload raw data to neurospeed)
    
    input:
        config - dictionary with config for this epxeriment program 
        hia_config dictionary
        
    output 
        hia_user client object instance
        sensor info dictionary
    
    '''
    user1_auth = Auth_AS_User_Handler(hia_config)
    user1_auth.login()
    
    hia_sensor_info_user1 = generate_sensor_info() 
     
    logging.debug('Generated sensor info: {}'.format(hia_sensor_info_user1) )
    hia_user1 = HIA_Client(user1_auth, hia_sensor_info_user1)   
    hia_user1.set_socket_connection_external_handler(connection_external_handler)
    hia_user1.set_socket_disconnect_external_handler(disconnect_external_handler)
    logging.debug("HIA_ID: " + str(user1_auth.get_hia_id()))   
    hia_user1._device_type = "EEG"
    hia_user1.connect()
    return hia_user1, hia_sensor_info_user1

    
    
 

if __name__ == '__main__':


    todays_date = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
    logfilename = "log"+todays_date+".txt"
    logfile = open(logfilename, "w")

        
    # datafolder, where to store the recorded data files locally:
    datafolder = "C:\\neurobrave"
    
    subject_name = "my_subject_1"
   
    hia_client = None 
    with open('config/hia_config.json') as f:
        hia_config = json.load(f)
 
    with open('config/dashboard_credentials.json') as f:
         customer_config = json.load(f)             
    # Authenticate as customer on neurospeed.io:
    print(f"using customer config {customer_config}")
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()

    
    recorder_id = macros.start_recording(customer_auth, hia_config, recorder_name = subject_name)
    
    

    max_hia_reconnect_retries = 3
    for reconnect_idx in range(max_hia_reconnect_retries):
        hia_client, hia_sensor_info = init_HIA_client(hia_config)
        if hia_client.is_connected():
            break
        else:
            logging.debug("HIA connection retrying...")
            time.sleep(3)
    
    if not hia_client.is_connected():
        logfile.write("FAILED connecting to cloud streaming!\n")
        sys.exit()
    


    # set the label for all the transmitted data from this point onwards:
    hia_client.set_label("experiment_part_1")
    
    # tranmsit some arbitrary data:
    datatype = "user_data"
    stream_id = list(filter(lambda x: x.startswith(datatype), list(hia_sensor_info.keys())))[0]

    for i in range(100):
        data = [[1,2,3,4,5,6,7,8,9,10], [1,2,3,4,5,6,7,8,9,10]]
        
        hia_client.send_data_direct(data, stream_id, send_without_timestamp=False, device_type=datatype)       
        time.sleep(0.2)
    
    
    # set the label for all the transmitted data from this point onwards:
    hia_client.set_label("experiment_part_2")



    for i in range(100):
        data = [[1,2,3,4,5,6,7,8,9,10], [1,2,3,4,5,6,7,8,9,10]]
        
        hia_client.send_data_direct(data, stream_id, send_without_timestamp=True, device_type=datatype)       
        time.sleep(0.2)
        

    # stop the data transmission:
        
    hia_client.disconnect()
    print("hia status: " + str(hia_client.is_connected()))   
    
    # stop the data recording on the cloud. Stopping the recording does not automatically download the data, another command does it.
    macros.stop_recording(recorder_id)
    

    # command the cloud engine to perform the data export and download the datafile:  
        
    # "exporter" finalizes the recording, converts it to human-readable .csv file and makes it available for download:
    # exporter name can be any string. exporter can export certain range of the recording, by selecting start and end timestamps. 
    # to export the entire recording, use: "start_timestamp_mode": "start", "end_timestamp_mode": "end"
 
    exporter_config =  {
    "custom_name": subject_name + "_exported",    
    "start_timestamp_mode": "start", 
    "end_timestamp_mode": "end"
    }
    
    save_folder =  os.path.join(os.getcwd(), datafolder)        
    exporter_filename = macros.download_data(customer_auth, recorder_id, exporter_config, hia_config, save_folder)
    
    
    logfile.close()
     
     



import time
from pathlib import Path
import os

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.downlink_room_as_user_handler import DownlinkRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
from neurospeed.hia_user_data.neurobrave_sensor_interface import HIA_Client

import threading


x= [12, 14]

my_custom_sensor_info = {
   'sensor_type': 'vehicle speed',
   'channel_count' : 2,
   'channel_map': ["ch1", "ch2", "timestamp"], # timestamp doesn't count as channel
   'manufacturer_id': 'toyota'

}

# see README(5)
def get_sensors_mockup():
    # generate SOME sensors stream ids
    user_data_stream_id =  "user_data" + '_' + UtilService.generateId(6)
    user_data_custom_stream_id = "some_custom_stream_id"

    sensor_info = dict()
    sensor_info[user_data_stream_id] =  {
        "device_type": "user_data",
        "channel_count": 2,
        "channel_map": ["ch_0", "ch1", "timestamp"], # timestamp doesn't count as channel
        "sensor_id" : user_data_stream_id,
        "stream_id": user_data_stream_id,
        "stream_state" : "enabled"
    } 
    
    sensor_info[user_data_custom_stream_id] =  {
        "device_type": "user_data",
        "channel_count": 2,
        "channel_map": ["ch_9",  "ch10", "timestamp"], # timestamp doesn't count as channel
        "sensor_id" : user_data_custom_stream_id,
        "stream_id": user_data_custom_stream_id,
        "stream_state" : "enabled"
    } 
    
    return sensor_info
 


def downlink_message_external_handler(downlink_handler_instance, payload):
    username = downlink_handler_instance.get_username()
    print('~~~external user', username,' downlink message: ', payload)


# send dummy data loop for some stream_id. must be activated after socket successful connection.
def generate_data(hia_client, stream_id):
    for i in range(5000):
        if hia_client.is_connected() == False: # stop sending data if disconnected
            raise ValueError("hia for " , hia_client.get_username(), " disconnected , stopping data generator thread")
        if hia_client.is_stream_enabled(stream_id): # send only if stream enabled for that sensor
            hia_client.send_data(x.copy(), stream_id)
        time.sleep(0.05)
        
    hia_client.disconnect() #disconnect after finish, just for the example


def disconnect_external_handler(hia_client):
   print("hia:[", hia_client.get_hia_id(),"]", "user:[", hia_client.get_username(),"] disconnected ")
    
 
    
# would be called for each hia after successfuly connection  
def connection_external_handler(hia_client):
    username = hia_client.get_username()
    hia_id = hia_client.get_hia_id()
    print('connected as ', username, " with hia: " ,hia_id, ". sending data..")
    
    # generate dummy data for each sensor inside HIA_Client, 
    sensor_info = hia_client.get_sensor_info()
    for stream_id in sensor_info:
        print("activating data generator thread for stream_id:", stream_id , " username: ", username, "hia_id: ", hia_id)
        user_data_data_generator = threading.Thread(target = generate_data, args=[hia_client, stream_id])
        time.sleep(0.1)
        user_data_data_generator.start()

    
def run_user_test_flow(config_path):
    # load user configuration
    user_config_path = os.path.join(config_path)
    user_config = UtilService.load_config_file(user_config_path)
    
    # create Auth_AS_User_Handler and pass user configuration 
    user_auth = Auth_AS_User_Handler(user_config)
    
    # login to NeuroSpeed as user
    user_auth.login()
    time.sleep(1)
    # get pre-defined sensor mockup information
    hia_sensor_info_user = get_sensors_mockup() 
    
    print('Generated mockup sensors for username: {}'.format(user_config["username"]))
    
    # create HIA instance and pass user's Auth_Handler along with generated sensors info
    hia_client = HIA_Client(user_auth, hia_sensor_info_user)
    
    # This is example for periodical sending of data from a thread
    
    # attach external handlers this external dis\connection handler would be called on dis\connection 
    hia_client.set_socket_connection_external_handler(connection_external_handler) 
    hia_client.set_socket_disconnect_external_handler(disconnect_external_handler)
    
    # connect via HIA_Client to NeuroSpeed Pipeline 
    hia_client.connect()
    ## once connection established, "connection_external_handler" function would be called and result in mockup data generation 
    
    
    # example connecting to SSR downlink and listen on events from customer
    user_downlink = DownlinkRoom_AS_User_Handler(user_auth)
    user_downlink.set_downlink_router_external_handler(downlink_message_external_handler)
    user_downlink.connect()
    
    
    # This example of simple manual data sender:
    time.sleep(5) 
    if hia_client.is_connected():
        first_stream_id = next(iter(hia_sensor_info_user)) # take first stream_id for the example
        hia_client.set_label("welcome")      
        time.sleep(5)
        hia_client.set_label("goodbye")
  
    hia_client.disconnect()
    
# example of login, connect and send data as 2 different HIA users with 2 mockup sensors for each
# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def run_HIAs():
    
    config_path_user1 = os.path.join(str(Path().absolute()) ,"config","hia_config.json")
    run_user_test_flow(config_path_user1)
    
  #  config_path_user2 = os.path.join(str(Path().absolute()) ,"config","hia_config2.json")
   # run_user_test_flow(config_path_user2)


if __name__ == '__main__':    
    run_HIAs()


# -*- coding: utf-8 -*-
"""
Created on Fri Sep 10 12:32:58 2021

@author: NeuroBrave LTD


this code is an example application that performs the following:
    
    1. connects to NeuroSpeed.io cloud with "customer level" authentication (customer level is admin level which is higher than user level)
    2. listens to user's connect/disconnect events under that customer
    3. when user connects, uses the SSR Downlink protocol to send an arbitrary message string to the newly connected machine.
    
"""

import time
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.api_http_handlers.gateway_http_handler import Gateway_HttpHandler 
from neurospeed.utils.helper_service import UtilService
from neurospeed.api_socket_handlers.customer_room_handler import CustomerRoom_Handler
from neurospeed.api_socket_handlers.downlink_room_as_customer_handler import DownlinkRoom_AS_Customer_Handler

global customer_auth


# listen on connected hia events and send downlink message to connected user machine
def external_hia_device_connection_handler(event):
    global customer_auth

    username = event["username"]
    hia_id = event["hia_id"]
    sensors = event["sensors"]
    stream_ids = list(sensors.keys())
    print('external_hia_device_connection_handler event from User: {}, HIA: {} Stream_ids: {}'.format(username ,hia_id, stream_ids ))
    
    # Example connect to user machine with SSR_Socket downlink. SSRsocket allows downlink between customer and user on any machine. 
    downlinkRoom = DownlinkRoom_AS_Customer_Handler(customer_auth, username)
    downlinkRoom.connect()
    time.sleep(3)
    if downlinkRoom.is_connected():
        message = "hello user {}".format(username) 
        print("sending downlink message from customer")
        downlinkRoom.route_message(message)
    
    time.sleep(2)
    downlinkRoom.disconnect()
       

def external_hia_device_disconnect_handler(event):
    global customer_auth

    username = event["username"]
    hia_id = event["hia_id"]
    sensors = event["sensors"]
    stream_ids = list(sensors.keys())
    print('external_user_device_disconnect_handler event from User: {}, HIA: {} Stream_ids: {}'.format(username ,hia_id, stream_ids ))

        

# Hint: you can set "Verbose_socket_log": "True" in customer config to enable more logging
def main():
    global customer_auth
    
    # load customer config file
    customer_config = UtilService.load_config_file('config/customer_config.json')
    
    # Authenticate as customer
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()

    # connect to 'CustomerRoom_Handler', and listen on events across all users
    customerRoom = CustomerRoom_Handler(customer_auth)
    customerRoom.set_hia_connection_external_handler(external_hia_device_connection_handler)
    customerRoom.set_hia_disconnect_external_handler(external_hia_device_disconnect_handler)
    customerRoom.connect()
 
    

if __name__ == '__main__':    
    main()  


#  copyright 2020 NeuroBrave LTD
#simple API for handling SSR's (sensor side reactor) events that are
# pushed from the NB client as a result of user's interaction with the processing routines



from pathlib import Path
import time
import os

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.downlink_room_as_user_handler import DownlinkRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService




def downlink_message_external_handler(downlink_handler_instance, payload):
    username = downlink_handler_instance.get_username()
    print('Recieved downlink message from {}'.format(username))

    

def main():
    
    # load users configuration
    # user1_config_path = os.path.join(str(Path().absolute()) ,"config","hia_config1.json")
    config_user1 = UtilService.load_config_file('config/hia_config1.json')

    # create Auth_AS_User_Handler for each user and pass user configuration 
    user1_auth = Auth_AS_User_Handler(config_user1)
    user1_auth.login()

    
    # example connecting to SSR downlink and listen on events from customer
    user1_downlink = DownlinkRoom_AS_User_Handler(user1_auth)
    user1_downlink.set_downlink_router_external_handler(downlink_message_external_handler)
    user1_downlink.connect()
    
    # send message to all user machines connected to this downlink room except for the sender itself and the Customer
    time.sleep(2)
    if user1_downlink.is_connected():
       print('sending message to user machines')
       message = "Hello to every USER machine connected to this downlink room.\n" 
       user1_downlink.route_message(message)
    
    


if __name__ == '__main__':    
    main()  



# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""

import time
from pathlib import Path
import os
from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
import socket
import json

from lib import send_email

global previous_message
global alert_message_sent
alert_message_sent = False
previous_message= None

global user_auth
bad_channel_statistics={}
def EEG_processing_handler1(payload):
    '''
    this function extracts data from payload each time a packet is received - typically 1 packet every second.
    Parameters
    ----------
    payload : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    example payload["output"] structure:
        {'heart_rate': {'estimated_heart_rate': 59.53488372093023}, 'ICA': {'artifact_detected': 0, 'artifact_power': -1}, 'bad_channel': [0, 0, 0, 0], 'engagement_upgraded_index_h-b': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_manf_index_h-p': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_upgraded_index_h-p': {'hard_decision': 2, 'soft_decision_v1': 2, 'soft_decision_v2': 2}, 'focus_attention_manf_index_h-b': {'hard_decision': 0, 'soft_decision_v1': 0, 'soft_decision_v2': 0}, 'engagement_upgraded_index_h-p': {'hard_decision': 0, 'soft_decision_v1': 0.012534856796264648, 'soft_decision_v2': 0.025069713592529297}, 'focus_attention_upgraded_index_h-b': {'hard_decision': 0, 'soft_decision_v1': 0, 'soft_decision_v2': 0}}

    example  payload["sensor_info"] structure: 
    
    '''
     

    # EEG_sensor_information = {}
    # EEG_sensor_information["EEG_channel_names"] = payload["sensor_info"]["channel_map"]
    # EEG_sensor_information["EEG_channel_num"] = len(payload["sensor_info"]["channel_map"])    
    # EEG_sensor_information["sampling_frequency"] = payload["sensor_info"]["sampling_frequency"]
    # print(EEG_sensor_information)

    print("got EEG packet")
    # print(payload["output"])
    
    # print(payload["output"]['bad_channel'])
      
def ppg_handler_func(payload):
    # print(payload["output"]["heart_rate_ppg"])
    pass
    

def EEG_processing_handler2(payload):
    # your implementation
    # print('EEG_processing_handler2')
    print(payload["output"].keys())
    pass
    
def customer_gamepad_msg_handler_function(payload):
    # your implementation
    print('customer_gamepad_msg_handler_function')
    pass

def customer_user_data_msg_handler_function(payload):
    # your implementation
    print('customer_user_data_msg_handler_function')
    pass


def report_missed_messages(email_msg = ''):
    print("message stream timout detected, sending alert")
    email_subj = "neurospeed disconnect alert"
    
    email_attachment_files = []
    # this function alerts user about broken data stream
    send_email.send_mail(send_from = 'oleg@neurobrave.com', send_to = 'oleg@neurobrave.com',
        subject = email_subj, message = email_msg, files=email_attachment_files,
        server = 'mail.smtp2go.com', port = 2525, username = "neurobrave", password="qibXXC603NzwZrFc")
    

def smartwatch_handler_func(payload):
    global previous_message
    global alert_message_sent
    alert_message_sent = False
    previous_message = time.time()
    print(payload["data"])
    
generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler1, EEG_processing_handler2],
    'gamepad': [customer_gamepad_msg_handler_function],
    'user_data': [customer_user_data_msg_handler_function],
    'ppg': [ppg_handler_func],
    'smartwatch': [smartwatch_handler_func]
}


def user_data_external_handler(payload):
    username = user_auth.get_username()
    stream_id = payload["stream_id"]
    device_type = payload["device_type"]
    hia_id = payload["hia_id"]
    sensor_info = payload["sensor_info"]
    # print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, hia_id, stream_id, device_type))
    print(".", end='')
    # execute each function handler for relevant device_type
    device_type = payload["device_type"].lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)  
    if payload["type"] =="disconnect": 
        report_missed_messages(f"client {payload['hia_id']} for user {payload['username']} disconnect from sensor detected!")
        
    


# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    global user_auth
    global previous_message
    global alert_message_sent
    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config\\","hia_config.json"))
    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    userRoom.set_device_events_external_handler(user_device_event_external_handler)

    userRoom.connect()

    previous_message = time.time()
    timeout = 15 #seconds
    last_printed_msg = previous_message
    alert_message_sent = False
    while True:
        if last_printed_msg != previous_message:
            print(f"last msg received {previous_message}")
            last_printed_msg = previous_message
        if time.time() - previous_message > timeout: 
            if not alert_message_sent:
                report_missed_messages("missed packets from sensor detected!")
                alert_message_sent = True
        time.sleep(0.5)

if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 14 13:42:26 2021

@author: NeuroBrave



this code is a working axample of how to operate the built-in raw data recorder at NeuroSpeed.io

the code performs 4 main tasks:
    
    1. authentication in the neurospeed.io system with user and password
    
    2. starting and stopping the recorder: 
        recorder is set up and records for 15 seconds,
        then it's paused for 5 seconds, 
        then reactivated for another 5 seconds, 
        then finally stopped.
        there's an option to delete the recorder after recording. Do not delete the recording before it's data has been exported using the exporter.
    
    3. running the data exporter (that converts the recording into downloadable .csv file)
    
    4. donwloading the recording file
    
    
    
    there are plenty of configuration options for the recorder, see the comments within the code.

"""

import time
from pathlib import Path
import os
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.api_http_handlers.recorder_http_handler import UserRoom_Recorder_Handler
from neurospeed.api_http_handlers.exporter_http_handler import UserRoom_Recorder_Exporter_Handler
from neurospeed.utils.helper_service import UtilService

import urllib.request

global customer_auth
global recorder_handler
global exporter_handler
global username


def run_recorder_flow():
    global recorder_handler
    global username
    # create recorder. recorder created with status "resource_pending". once available recording resources found and assgined
    # recorder status will change to "pending", then once recorder picked by assigned recorder resources and start to record status will change to "recording"

    recorder = recorder_handler.create_recorder(username)
    recorder_id = str(recorder["id"])
    print('created recorder with id', recorder_id)
    
    recorder = recorder_handler.get_recorder(recorder_id)     
    print("recorder status: " + recorder["status"])
    
    recorder_recording = False
    max_query_attemps = 10
    query_attemps = 1
    while (recorder_recording != True & (query_attemps < max_query_attemps) ):
        recorder = recorder_handler.get_recorder(recorder_id)   
        print("recorder status: " + recorder["status"])
        if (recorder["status"] == "recording"):
            print('recorder started recording')
            recorder_recording = True
        else:
            print('recorder not recording yet')
            query_attemps = query_attemps + 1
            time.sleep(3)
             
    if (recorder_recording != True):
        print('recorder not yet recording after ' + str(query_attemps)  + ' query attempts')
        recorder_handler.delete_recorder(recorder_id)
        return
    
    
    print('~~~~~~~~~~~~~~~~~~')
    
    recording_duration = 20
    time.sleep(recording_duration-5)

    #once recorder starts recording, you can change it's status to paused or stopped. you can resume paused recorder but not stopped recorder
    # pause recorder for 5 seconds:
    res = recorder_handler.update_recorder(recorder_id, "paused")
    recorder = recorder_handler.get_recorder(recorder_id)     
    print("recorder status: " + recorder["status"])
    time.sleep(5)
    
    # reactivate paused recorder and add another 5 seconds to the recording
    recorder_handler.update_recorder(recorder_id, "pending") 
    recorder = recorder_handler.get_recorder(recorder_id)     
    print("recorder status: " + recorder["status"])
    
    time.sleep(5)
    
    recorder_handler.update_recorder(recorder_id, "pending") 
    recorder = recorder_handler.get_recorder(recorder_id)     
    print("recorder status: " + recorder["status"])
    
    
    recorder_handler.update_recorder(recorder_id, "stopped") 
    #recorder_handler.delete_recorder(recorder_id) # uncomment this to delete recorder
    recorders = recorder_handler.list_recorders(username)
    print(recorders)
    
    return recorder_id


def run_exporter_flow(recorder_id, exporter_config):
    global exporter_handler
    
    # create exporter, if recorder has no data, request will fail.
    exporter = exporter_handler.create_exporter(recorder_id, exporter_config)
    exporter_id = str(exporter["id"])
    
    # once exporter created, query it's status until it "exported".
    exporter_exported = False
    max_query_attemps = 20
    query_attemps = 1
    while (exporter_exported != True & (query_attemps < max_query_attemps) ):
        exporter = exporter_handler.get_exporter(exporter_id)   
        print("exporter status: " + exporter["status"])
        if (exporter["status"] == "exported"):
            print('exporter finished exporting')
            exporter_exported = True
        else:
            print('exporter not yet exported the files')
            query_attemps = query_attemps + 1
            time.sleep(10)
             
    if (exporter_exported != True):
        print('exporter not yet exported the file after  ' + str(query_attemps)  + ' query attempts')
        exporter_handler.delete_exporter(exporter_id)
        return
   
    # once exporter exported the file, you can ask for an URL to download the exported file. 
    #the url available for only short time before it deprecates
    url_response = exporter_handler.get_exported_url(exporter_id)
    
    
    time.sleep(2)
   #delete_exporter(exporter_id) # uncomment this to delete exporter
    time.sleep(1)
    # list all exporters under username
    exporters = exporter_handler.list_exporters(username)
    print(exporters)
    
    return url_response
    
def main():
    global customer_auth
    global recorder_handler
    global exporter_handler
    global username

    # load Customer configuration
    customer_config_path = os.path.join(str(Path().absolute()) ,"config","customer_config.json")
    customer_config = UtilService.load_config_file(customer_config_path)
    
    # Authenticate as customer
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()
    
    os.path.join(str(Path().absolute()) ,"neurospeed\\config","hia_config1.json")
    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config","hia_config1.json"))
    username = config_user1["username"] # or set here another custom username



    # run data recorder: 
    
    recorder_handler = UserRoom_Recorder_Handler(customer_auth)
    recorder_id = run_recorder_flow()
    print('RECORDER ID: {}'.format(recorder_id))



    # run exporter of recorder data into downloadable .csv format: 
    exporter_handler = UserRoom_Recorder_Exporter_Handler(customer_auth)

    # exporter configuration
    exporter_config =  {
        # name your exporter up to 128 chars
        "custom_name": "export me gently", 
        
         # start_timestamp_mode. values options: 
         # 1) "start" - from the first recorded timestamp.  
         # 2) "custom_value" - custom starting timestamp in seconds(!) . 
         # for example : "1629900352" or for better Accuracy "1629900352.11..." 
         # but not 162990035211" which is timestamp represented in miliseconds instead of seconds!
        "start_timestamp_mode": "start", 
        
         #  custom_start_timestamp required if "start_timestamp_mode" is "custom_value"
        # "custom_start_timestamp": 1629900352.11111, 
        
        # end_timestamp_mode. values options:
        # 1) "end" - until the last recorded timestamp  
        # 2) "custom_value" - custom end timestamp in seconds(!). same logic as "start_timestamp_mode"
        # 3) interval_minutes - interval in minutes. set
        "end_timestamp_mode": "end", 
        
        # custom_timestamp_interval required if "end_timestamp_mode" is "interval_minutes", 
        #"custom_timestamp_interval": 3 ,
        
        # custom_end_timestamp required if "end_timestamp_mode" is "custom_value", 
       # "custom_end_timestamp": "1629900353.11111", 
       
       # to exporting specific streams, add their stream_id into the array.  
       # to exporting all recorded streams skip this configuration
       # "stream_ids": ["stream_id1", "stream_id2"]
    }
    download_url = run_exporter_flow(recorder_id, exporter_config) 

    try: 
        savefilename = "recording_save.csv"
        urllib.request.urlretrieve(download_url, savefilename)
    except: 
        print("couldn't download file")

if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
Created on Fri Sep 18 11:04:53 2021

@author: NeuroBrave
"""

from pathlib import Path
import os
from neurospeed.utils.helper_service import UtilService

from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.api_http_handlers.gateway_http_handler import Gateway_HttpHandler 

global customer_auth

 
#### Gateway_HttpHandler.get_connections response: (TypeScript syntax)
#
# {
#    "pager": {
#        "total_records": number           # total records\entries found applying search filters
#        "total_pages": number             # total number of entries after applying search filters
#        "page_number": number             # requested page number
#        "page_size": number               # requested page number 
#        "actual_page_size": number        # it might be that acutal returned page size is smaller then requested. actual_page_size === items.length
#        "items": any[]                    # list of found items, (list of Connection_Dto objects in this case)
#     },
#   "filters": {}                          # applied filters
# }
#
################################# 
# Connection_Dto {
#    id: number;
#    username: string;
#    sensors: any[]
#    socket_id: string;
#    sensors_count: number
#    online: {hours: number, minutes: number}
#    hia_id: string;
#    is_connected: boolean;
#    connected_at: Date;
#    disconnected_at: Date;
#    updated_at: Date;
# }
####

def print_list_response(list_response): 
    pager = list_response["pager"]
    
    total_records = pager["total_records"] 
    total_pages = pager["total_pages"]
    page_number = pager["page_number"]
    page_size = pager["page_size"]    
    actual_page_size = pager["actual_page_size"]  
    items = pager["items"]
    
    filters = list_response["filters"]  

    
    print("HIA_CONNECTION - query with filters: {} results in {} entries and {} pages".format(filters, total_records, total_pages))
    print("HIA_CONNECTION - page_number: {} requested page_size: {} actual page_size: {} ".format(page_number, page_size, actual_page_size))
    
    for item in items:
        # custom
        pass


# Hint: you can set "Verbose_socket_log": "True" in customer config to enable more logging
def main():
    global customer_auth

    # load customer config file
    customer_config_path = os.path.join(str(Path().absolute()) ,"config","customer_config.json")
    customer_config = UtilService.load_config_file(customer_config_path)
    
    # Authenticate as customer
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()
    
    gateway = Gateway_HttpHandler(customer_auth)
    connected_users = gateway.get_connected_users()

    username_array = []
    for user in connected_users:
        username_array.append(user["username"])
        
    print('connected user array: {}'.format(username_array))
    
    # using get_connections() API call, you can query user HIA device connections
    # returns list of connection entries, already sorted where recent updated entires come first
    if len(username_array ) > 0:
        print('quering Neurospeed API for connected user device')

        page_number = 1 # required
        page_size = 50 # required. Max page size: 100
        
        pager = { # required
            "page_number": page_number,
            "page_size": page_size
        }
        filters = { # not required
            "is_connected": True, # if is_connected is not set or False, query would return both connected and disconnected entries
            "usernames": username_array, # array of usernames, if array is empty or not set, query would search for entires across all users
        }
        
        connected_connections_response = gateway.get_connections(pager, filters)
        print_list_response(connected_connections_response)
    else: 
        print('No connected users')

    # example query all connections (currently connected and disconnected) for all users:
    print('Quering Neurospeed API for all user device connections')   
    
    page_number = 1 
    page_size = 20 


    pager = {
        "page_number": page_number,
        "page_size": page_size
    }
    filters = {}
    
    all_connections_response = gateway.get_connections(pager, filters )
    print_list_response(all_connections_response)
   




if __name__ == '__main__':    
    main()  
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 14 13:42:26 2021

@author: NeuroBrave



this code is a working axample of how to operate the built-in raw data recorder at NeuroSpeed.io

the code performs 4 main tasks:
    
    1. authentication in the neurospeed.io system with user and password
    
    2. starting and stopping the recorder: 
    
    3. running the data exporter (that converts the recording into downloadable .csv file)
    
    4. donwloading the recording file
    
    
    
    there are plenty of configuration options for the recorder, see the comments within the code.

"""


from pathlib import Path
import os
from neurospeed.auth.auth_as_customer_handler import Auth_AS_Customer_handler
from neurospeed.utils.helper_service import UtilService
from neurospeed.macros import macros

global customer_auth
global username


    
def main():
    global customer_auth
    global recorder_handler
    global exporter_handler
    global username

    # load Customer configuration
    customer_config_path = os.path.join(str(Path().absolute()) ,"config","customer_config.json")
    customer_config = UtilService.load_config_file(customer_config_path)
    
    # Authenticate as customer
    customer_auth = Auth_AS_Customer_handler(customer_config)
    customer_auth.login()
    
    os.path.join(str(Path().absolute()) ,"neurospeed\\config","hia_config1.json")
    config_user1 = UtilService.load_config_file(os.path.join(str(Path().absolute()) ,"config","hia_config1.json"))
    username = config_user1["username"] # or set here another custom username



    # run data recorder: 
    recorder_name="mydata"
    recorder_id = macros.start_recording(customer_auth, config_user1, recorder_name)



    # ****************************************************** 
    # send some data here, empty recorders are not generating downloaded file!
    


    # run exporter of recorder data into downloadable .csv format: 
    # exporter configuration
    exporter_config =  {
        # name your exporter up to 128 chars
        "custom_name": "export me gently", 
        
         # start_timestamp_mode. values options: 
         # 1) "start" - from the first recorded timestamp.  
         # 2) "custom_value" - custom starting timestamp in seconds(!) . 
         # for example : "1629900352" or for better Accuracy "1629900352.11..." 
         # but not 162990035211" which is timestamp represented in miliseconds instead of seconds!
        "start_timestamp_mode": "start", 
        
         #  custom_start_timestamp required if "start_timestamp_mode" is "custom_value"
        # "custom_start_timestamp": 1629900352.11111, 
        
        # end_timestamp_mode. values options:
        # 1) "end" - until the last recorded timestamp  
        # 2) "custom_value" - custom end timestamp in seconds(!). same logic as "start_timestamp_mode"
        # 3) interval_minutes - interval in minutes. set
        "end_timestamp_mode": "end", 
        
        # custom_timestamp_interval required if "end_timestamp_mode" is "interval_minutes", 
        #"custom_timestamp_interval": 3 ,
        
        # custom_end_timestamp required if "end_timestamp_mode" is "custom_value", 
       # "custom_end_timestamp": "1629900353.11111", 
       
       # to exporting specific streams, add their stream_id into the array.  
       # to exporting all recorded streams skip this configuration
       # "stream_ids": ["stream_id1", "stream_id2"]
    }
    save_folder = os.getcwd()
    download_filepath = macros.download_data(customer_auth, recorder_id, exporter_config, config_user1, save_folder)
    print(f"downloaded file: {download_filepath}")

if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""


GRAPH_ENGAGEMENT = False

import rtmidi
import re
import sys
from pathlib import Path
import os

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService

global user_auth

import statistics
import queue
import time
from vispy import gloo, app, visuals
import numpy as np
import math
from seaborn import color_palette
from scipy.signal import lfilter, lfilter_zi
from mne.filter import create_filter

import socket
import json



global brain_wave_power_max
global midiout
brain_wave_power_max = 0




stream_q_EEG = queue.Queue(-1)


VERT_SHADER = """
#version 120
// y coordinate of the position.
attribute float a_position;
// row, col, and time index.
attribute vec3 a_index;
varying vec3 v_index;
// 2D scaling factor (zooming).
uniform vec2 u_scale;
// Size of the table.
uniform vec2 u_size;
// Number of samples per signal.
uniform float u_n;
// Color.
attribute vec3 a_color;
varying vec4 v_color;
// Varying variables used for clipping in the fragment shader.
varying vec2 v_position;
varying vec4 v_ab;
void main() {
    float n_rows = u_size.x;
    float n_cols = u_size.y;
    // Compute the x coordinate from the time index.
    float x = -1 + 2*a_index.z / (u_n-1);
    vec2 position = vec2(x - (1 - 1 / u_scale.x), a_position);
    // Find the affine transformation for the subplots.
    vec2 a = vec2(1./n_cols, 1./n_rows)*.9;
    vec2 b = vec2(-1 + 2*(a_index.x+.5) / n_cols,
                    -1 + 2*(a_index.y+.5) / n_rows);
    // Apply the static subplot transformation + scaling.
    gl_Position = vec4(a*u_scale*position+b, 0.0, 1.0);
    v_color = vec4(a_color, 1.);
    v_index = a_index;
    // For clipping test in the fragment shader.
    v_position = gl_Position.xy;
    v_ab = vec4(a, b);
}
"""

FRAG_SHADER = """
#version 120
varying vec4 v_color;
varying vec3 v_index;
varying vec2 v_position;
varying vec4 v_ab;
void main() {
    gl_FragColor = v_color;
    // Discard the fragments between the signals (emulate glMultiDrawArrays).
    if ((fract(v_index.x) > 0.) || (fract(v_index.y) > 0.))
        discard;
    // Clipping test.
    vec2 test = abs((v_position.xy-v_ab.zw)/v_ab.xy);
    if ((test.x > 1))
        discard;
}
"""


global UPDATE_TIMER
UPDATE_TIMER = 1.5

window_title = "use mouse wheel for vertical zoom, +/- keys for horizontal zoom. D key toggles high pass filter"

class Canvas(app.Canvas):
    def __init__(self, stream_info, scale=1000, filt=True):
        app.Canvas.__init__(self, title=window_title, keys='interactive')   
        window = 10
        
        self.sfreq = stream_info['srate']
        self.n_chans = stream_info['n_chans']
        ch_names = stream_info['ch_names']
        
        
        n_samples = int(self.sfreq * window)


        # Number of cols and rows in the table.
        n_rows = self.n_chans
        n_cols = 1

        # Number of signals.
        m = n_rows * n_cols

        # Number of samples per signal.
        n = n_samples

        # Various signal amplitudes.
        amplitudes = np.zeros((m, n)).astype(np.float32)
        # gamma = np.ones((m, n)).astype(np.float32)
        # Generate the signals as a (m, n) array.
        y = amplitudes

        color = color_palette("RdBu_r", n_rows)

        color = np.repeat(color, n, axis=0).astype(np.float32)
        # Signal 2D index of each vertex (row and col) and x-index (sample index
        # within each signal).
        index = np.c_[np.repeat(np.repeat(np.arange(n_cols), n_rows), n),
                      np.repeat(np.tile(np.arange(n_rows), n_cols), n),
                      np.tile(np.arange(n), m)].astype(np.float32)

        self.program = gloo.Program(VERT_SHADER, FRAG_SHADER)
        self.program['a_position'] = y.reshape(-1, 1)
        self.program['a_color'] = color
        self.program['a_index'] = index
        self.program['u_scale'] = (1., 1.)
        self.program['u_size'] = (n_rows, n_cols)
        self.program['u_n'] = n

        # text
        self.font_size = 24.
        self.names = []
        self.quality = []
        for ii in range(self.n_chans):
            text = visuals.TextVisual(ch_names[ii], bold=True, color='white')
            self.names.append(text)
            text = visuals.TextVisual('', bold=True, color='white')
            self.quality.append(text)

        self.quality_colors = color_palette("RdYlGn", 11)[::-1]

        self.scale = scale
        self.n_samples = n_samples
        self.filt = filt
        self.af = [1.0]

        self.data_f = np.zeros((n_samples, self.n_chans))
        self.data = np.zeros((n_samples, self.n_chans))

        self.bf = create_filter(self.data_f.T, self.sfreq, None, 10.,
                                method='fir')

        zi = lfilter_zi(self.bf, self.af)
        self.filt_state = np.tile(zi, (self.n_chans, 1)).transpose()

        self._timer = app.Timer(UPDATE_TIMER, connect=self.on_timer, start=True)
        
        gloo.set_viewport(0, 0, *self.physical_size)
        gloo.set_state(clear_color='black', blend=True,
                       blend_func=('src_alpha', 'one_minus_src_alpha'))
        self.show()

    def on_key_press(self, event):

        # toggle filtering
        if event.key.name == 'D':
            self.filt = not self.filt

        # increase time scale
        if event.key.name in ['+', '-']:
            if event.key.name == '+':
                dx = -0.05
            else:
                dx = 0.05
            scale_x, scale_y = self.program['u_scale']
            scale_x_new, scale_y_new = (scale_x * math.exp(1.0 * dx),
                                        scale_y * math.exp(0.0 * dx))
            self.program['u_scale'] = (
                max(1, scale_x_new), max(1, scale_y_new))
            self.update()

    def on_mouse_wheel(self, event):
        dx = np.sign(event.delta[1]) * .05
        scale_x, scale_y = self.program['u_scale']
        scale_x_new, scale_y_new = (scale_x * math.exp(0.0 * dx),
                                    scale_y * math.exp(2.0 * dx))
        self.program['u_scale'] = (max(1, scale_x_new), max(0.01, scale_y_new))
        self.update()

    def on_timer(self, event):
        """Add some data at the end of each signal (real-time signals)."""       
        samples = stream_q_EEG.get(timeout = 1)
            # samples = [[-1 for j in range(self.n_chans)  ] for i in range(int(self.sfreq * UPDATE_TIMER))]      
        
        if True:
            samples = np.array(samples)[:, ::-1]

            self.data = np.vstack([self.data, samples])
            self.data = self.data[-self.n_samples:]
            filt_samples, self.filt_state = lfilter(self.bf, self.af, samples,
                                                    axis=0, zi=self.filt_state)
            self.data_f = np.vstack([self.data_f, filt_samples])
            self.data_f = self.data_f[-self.n_samples:]

            if self.filt:
                plot_data = self.data_f / self.scale
            elif not self.filt:
                plot_data = (self.data - self.data.mean(axis=0)) / self.scale

            sd = np.std(plot_data[-int(self.sfreq):],
                        axis=0)[::-1] * self.scale
            co = np.int32(np.tanh((sd - 30) / 15) * 5 + 5)
            for ii in range(self.n_chans):
                self.quality[ii].text = '%.2f' % (sd[ii])
                self.quality[ii].color = self.quality_colors[co[ii]]
                self.quality[ii].font_size = 12 + co[ii]

                self.names[ii].font_size = 12 + co[ii]
                self.names[ii].color = self.quality_colors[co[ii]]

            self.program['a_position'].set_data(
                plot_data.T.ravel().astype(np.float32))
            self.update()

    def on_resize(self, event):
        # Set canvas viewport and reconfigure visual transforms to match.
        vp = (0, 0, self.physical_size[0], self.physical_size[1])
        self.context.set_viewport(*vp)

        for ii, t in enumerate(self.names):
            t.transforms.configure(canvas=self, viewport=vp)
            t.pos = (self.size[0] * 0.025,
                     ((ii + 0.5) / self.n_chans) * self.size[1])

        for ii, t in enumerate(self.quality):
            t.transforms.configure(canvas=self, viewport=vp)
            t.pos = (self.size[0] * 0.975,
                     ((ii + 0.5) / self.n_chans) * self.size[1])

    def on_draw(self, event):
        gloo.clear()
        gloo.set_viewport(0, 0, *self.physical_size)
        self.program.draw('line_strip')
        [t.draw() for t in self.names + self.quality]


global EEG_sensor_information
EEG_sensor_information=None

def EEG_processing_handler1(payload):
    
    # global midiout
    global brain_wave_power_max
    global EEG_sensor_information    
    global midiout
    if EEG_sensor_information == None:
        EEG_sensor_information = {}
        EEG_sensor_information["EEG_channel_names"] = payload["sensor_info"]["channel_map"]
        EEG_sensor_information["EEG_channel_num"] = len(payload["sensor_info"]["channel_map"])    
        EEG_sensor_information["sampling_frequency"] = payload["sensor_info"]["sampling_frequency"]


    # electrode_select is a list of EEG electrodes to be included in the calculation of the brainwave output. 
    # if a single electrode is desired, create a list with a single value:
    #  if all available electrodes are to be included, use this:
    # electrode_select = "all"
    electrode_select = [0]

    # print(payload["output"]['brainwave_power'])
    
    if "all" in electrode_select:
        brain_wave_power = statistics.mean(payload["output"]['brainwave_power']['alpha'])
    else:
        brain_wave_power = statistics.mean([payload["output"]['brainwave_power']['alpha'][electrode_index] for electrode_index in electrode_select])
    
    if brain_wave_power > brain_wave_power_max:
        brain_wave_power_max = brain_wave_power
    
    if brain_wave_power_max > 0:
        output_value = int(127 * brain_wave_power/brain_wave_power_max)      
    else:
        output_value = 0
    
    if output_value >127:
        output_value = 127
    midi_Control_Change_number = 7
    print("sent CC value: ", output_value)
        
    midiout.send_message([88, 11, 64])
    # midiout.send_message([88,midi_Control_Change_number,output_value])


    
    if GRAPH_ENGAGEMENT: 
        values = []
        for i in range(100):
            values.append([brain_wave_power])
        stream_q_EEG.put(values, timeout = 1)

  


def EEG_processing_handler2(payload):
    # your implementation
    # print('EEG_processing_handler2')
    pass
    
def customer_gamepad_msg_handler_function(payload):
    # your implementation
    print('customer_gamepad_msg_handler_function')
    pass

def customer_user_data_msg_handler_function(payload):
    # your implementation
    print('customer_user_data_msg_handler_function')
    pass

generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler1, EEG_processing_handler2],
    'gamepad': [customer_gamepad_msg_handler_function],
    'user_data': [customer_user_data_msg_handler_function]
}



def user_data_external_handler(payload):
    username = user_auth.get_username()
    stream_id = payload["stream_id"]
    device_type = payload["device_type"]
    hia_id = payload["hia_id"]
    sensor_info = payload["sensor_info"]
    # print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, hia_id, stream_id, device_type))
    
    # execute each function handler for relevant device_type
    device_type = device_type.lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)
    pass;




midiout = rtmidi.MidiOut()
available_ports = midiout.get_ports()
if available_ports:
    connected = False
    while not connected:
        print("Available MIDI output ports in the system: ")
        print(available_ports)
        print("Select port by typing  in port number")
        port_number = int(re.findall('[0-9]+', input())[0])
        if port_number in range(0, len(available_ports)):
            print("selected port ",available_ports[port_number] , " opening...")
            midiout.open_port(port_number)
            connected = True    
            time.sleep(3)
            
        else:
            print('incorrect input, try again:')
else:
    print("no MIDI outputs found in system, terminating the program.")
    del midiout
    sys.exit()
    
# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    global user_auth
   

    user1_config_path = os.path.join(str(Path().absolute()) ,"config\\","hia_config1.json")
    print(user1_config_path)
    config_user1 = UtilService.load_config_file(user1_config_path)


    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    userRoom.set_device_events_external_handler(user_device_event_external_handler)

    userRoom.connect()
    



    time.sleep(5)
    print(EEG_sensor_information)
    
    if GRAPH_ENGAGEMENT: 
        stream_info = {'n_chans': 1,'srate': 100  , 'ch_names': ["engamement"]    }
        Canvas(stream_info)
        app.run()



    

if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
Created on Sun May 21 11:24:03 2023

@author: neurobrave
"""

import socket
import json

UDP_IP = "127.0.0.1"
UDP_PORT = 5005

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
sock.bind((UDP_IP, UDP_PORT))
while True:
    data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
    value = json.loads(data,'utf-8')["relaxation"]
    print("received message: %s" % data)
    print("received message: %s" % data)
    


# -*- coding: utf-8 -*-
"""
Created on Sun Sep 12 10:13:39 2021

@author: NeuroBrave
"""

import json
from neurospeed.utils.http_service import HttpService


# SSR => send events to HIA devices (enable\disable stream state, downlink etc..)
class SSR_HttpHandler:
    
    def __init__(self, customer_auth_handler):
        # create ssr http api instance
        self._ssr_api = self.SSR_HttpApi(customer_auth_handler)
    

    def change_stream_state(self, username, hia_id, stream_id, stream_state):
        print("Sending stream_state message to User: {} HIA: {} Stream: {} state: {}".format(username, hia_id, stream_id, stream_state))
        stream_state_payload = {
            "username": username, # required
            "hia_id": hia_id, # required
            "stream_id": stream_id, # required. take first sensor just for the example
            "stream_state": stream_state # required. options: 'disabled' or 'enabled'
        }     
    
        stream_state_response = self._ssr_api.change_stream_state(stream_state_payload)
        print("Stream state http response: {}".format(stream_state_response)  )
    
    def send_downlink(self, username, hia_id, payload):
        print("Sending downlink message to User: {} HIA: {}".format(username, hia_id))
        downlink_payload = {
            "username": username, # required
            "hia_id": hia_id,  # required
            "payload": json.dumps(payload) # required. payload is JSON object and can be anything up to 64 killobytes ~ 
        }     
    
        stream_state_response = self._ssr_api.downlink(downlink_payload)
        print("Downlink http response: {} ".format(stream_state_response) )
        
        
    class SSR_HttpApi:
    
        def __init__(self, auth_handler):
            self._http_service = HttpService()
            self._headers =  {
                "Authorization": "Bearer " + auth_handler.get_access_token()
            }
    
    
        def change_stream_state(self, payload):
            endpoint = '/gateway/ssr/stream_state'
           
            response = self._http_service.POST_request(endpoint, payload, self._headers)
            return response
                
                
        def downlink(self, payload):
            endpoint = '/gateway/ssr/downlink'
           
            response = self._http_service.POST_request(endpoint, payload, self._headers)
            return response


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 13 12:42:25 2021

@author: NeuroBrave
"""

import socketio 
from neurospeed.utils.api_config_service import ApiConfig

#target_is_downlink_room_as_customer
class DownlinkRoom_AS_Customer_Handler:
    
    def __init__(self, customer_auth_handler, username):

        self._auth_handler = customer_auth_handler
        self._customer_email = self._auth_handler.get_customer_email()
        self._username = username
        self._connection_status = False
        # Create UserTargetRoom and pass customer configuration and username 
        self._downlink_api = self.DownlinkRoom_Api(self._auth_handler, self._username)
        
    def downlinkRoom_socket_connection_handler(self):
        print("Customer {} connected to user's {} DownlinkRoom".format(self._customer_email, self._username ))
        self._connection_status = True
    
    
    def downlinkRoom_socket_disconnect_handler(self):
        print("Customer {} disconnected from user's {} DownlinkRoom".format(self._customer_email, self._username ))
        self._connection_status = False
        
    def route_message(self, payload):
        self._downlink_api.route_message(payload)
        
    
    # do socket connection 
    def connect(self):
        # attach relevant handlers for socket.io events
        self._downlink_api.set_connection_handler(self.downlinkRoom_socket_connection_handler)
        self._downlink_api.set_disconnect_handler(self.downlinkRoom_socket_disconnect_handler)

        self._downlink_api.connect()
        
        
    def disconnect(self):
        self._downlink_api.disconnect()
        
    def is_connected(self):
        return self._connection_status
            
    
    class DownlinkRoom_Api:
    
        def __init__(self, auth_handler, username):
            api_config = ApiConfig()
            self._socket_url = api_config.get_socket_url()
            
            self._access_token = auth_handler.get_access_token()
            self._username = username
            
            logger_on = True
            if auth_handler.is_verbose_log() == False:
                logger_on = False
            self._sio = socketio.Client(logger=logger_on, engineio_logger=False, reconnection_delay  = 5, reconnection = True) 
            
            
            
        def set_connection_handler(self, handler):
            self._sio.on('connect',handler = handler)
            
        def set_disconnect_handler(self, handler):
            self._sio.on('disconnect', handler = handler)
            
        def route_message(self, payload):
            self._sio.emit('router', payload)
         
            
        def connect(self):
           headers = {
               "jwt_token": self._access_token,
               "username": self._username, 
           }
        
           print("Downlink_Api - Connecting to user's {} Downlink room as customer ".format(self._username))
           self._sio.connect(url = self._socket_url, transports ='websocket', headers=headers, socketio_path= '/target_is_downlink_room_as_customer' ) 
    
        def disconnect(self):
           self._sio.disconnect()


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 14 14:02:51 2021

@author: NeuroBrave
"""

import json
import copy
from neurospeed.utils.http_service import HttpService


class UserRoom_Recorder_Exporter_Handler:
    
    def __init__(self, customer_auth_handler):
        # create exporter http api instance
        self._exporter_api = self.Exporter_HttpApi(customer_auth_handler)
        
    def create_exporter(self, recorder_id, config):
        create_exporter_payload = {
            "recorder_id": recorder_id,
            "exporter_config": config,
            "unlimited": "true"
        }
        payload_json = json.dumps(create_exporter_payload)
        response = self._exporter_api.create_exporter(payload_json)
        
        return response
    
    def get_exported_url(self, exporter_id):
        params  = {
            "exporter_id": exporter_id
        }
        response = self._exporter_api.get_exported_url(params)
        
        return response
    
    def list_exporters(self, username, page_number, page_size):
        payload  = {
            "pager": {
                "page_number": page_number,
                "page_size":page_size
            },
            "filters": {
                "username": username
            }
        }
        response = self._exporter_api.list_exporters(payload)
        return response
    
    def get_exporter(self, exporter_id):
        response = self._exporter_api.get_exporter(exporter_id)
        
        return response
        
    def delete_exporter(self, exporter_id):
        params  = {
            "exporter_id": exporter_id
        }
        response = self._exporter_api.delete_exporter(params)
        return response
        
    class Exporter_HttpApi:
        
        def __init__(self, auth_handler):
            # create http service instance
            self._exporter_endpoint = '/gateway/exporter'
            self._http_service = HttpService()
            self._headers =  {
                "Authorization": "Bearer " + auth_handler.get_access_token()
            }
            
        def get_headers(self):
            return copy.deepcopy(self._headers)
        
        def create_exporter(self, payload):
            endpoint = self._exporter_endpoint
            request_headers = self.get_headers()
            request_headers["content-type"] = "application/json"
            response = self._http_service.POST_request(endpoint, payload, request_headers )
            
            return response
        
        def list_exporters(self, payload ):
          endpoint = self._exporter_endpoint + "/list"
          
          request_headers = self.get_headers()
          request_headers["content-type"] = "application/json"
          response = self._http_service.POST_request(endpoint, json.dumps(payload), request_headers)
          return response
    
        def get_exported_url(self, params):
            endpoint = self._exporter_endpoint + "/url"

            response = self._http_service.GET_request(endpoint, params , self._headers)
            
            return response
        
        def get_exporter(self, exporter_id):
            endpoint = self._exporter_endpoint + "/" + str(exporter_id)
            response = self._http_service.GET_request(endpoint, {}, self._headers)
            
            return response
            
        def delete_exporter(self, params):
            endpoint = self._exporter_endpoint
            response = self._http_service.DELETE_request(endpoint, params, self._headers)
            
            return response


'''
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
import time

driver = webdriver.Chrome("C:\\author_data\\chromedriver\\95\\chromedriver.exe")  # Optional argument, if not specified will search path.
driver.get("https://web.whatsapp.com/")
wait = WebDriverWait(driver, 600)
# Replace 'Friend's Name' with the name of your friend
# or the name of a group
target = '"You"'
# Replace the below string with your own message
message = "Message sent using Python!!!"
 
x_arg = '//span[contains(@title,' + target + ')]'
group_title = wait.until(EC.presence_of_element_located((
    By.XPATH, x_arg)))
group_title.click()
inp_xpath = '//div[@class="_13NKt copyable-text selectable-text"][@data-tab="9"]'
input_box = wait.until(EC.presence_of_element_located((
    By.XPATH, inp_xpath)))

input_box.send_keys(message + Keys.ENTER)
driver.quit()

'''



# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""


GRAPH_ENGAGEMENT = True

from pathlib import Path
import os

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService
import time
import pywhatkit as whatsapp
import keyboard

global user_auth
global is_tab_open
is_tab_open = False

global EEG_sensor_information
global first_disctraction_time
global distraction_count
distraction_count = 0

EEG_sensor_information=None

def send_whatsapp(message):
    global is_tab_open    
    if not is_tab_open:
        whatsapp.sendwhatmsg_instantly("+972545862021", message, wait_time=10)
        is_tab_open=True
        time.sleep(0.1)
    else:
        # using whatsapp toolkit only required for the first time to open the browser window
        # after that we're typing and sending the message manually to make it quiker. 
        # please note that the browswer window must remain active for the script to work. 
        keyboard.write(message)
        time.sleep(0.2)
        keyboard.press_and_release('enter')
    print("is tab open? ", is_tab_open)

def EEG_processing_handler1(payload):
    
    '''
    this function is being called every time a packet arrives, depending on confifguration of the HIA source; usually this is once per second.
    the code extracts the useful sensor information that comes with every packet, but does it just one time to be elegant and not keep overwriting the EEG_sensor_information variable   
    
    # here some arbitrary logic is implemented as an example: after 5 distraction alerts 
    # within 1 minute, a WhatsApp notification message is dispatched
    
    thanks to Neurospeed, the implementation of distraction reminted is  just few lines of code. 
       
     '''  
         
    global EEG_sensor_information    
    global distraction_count
    global first_disctraction_time
    if EEG_sensor_information == None:
        EEG_sensor_information = {}
        EEG_sensor_information["EEG_channel_names"] = payload["sensor_info"]["channel_map"]
        EEG_sensor_information["EEG_channel_num"] = len(payload["sensor_info"]["channel_map"])    
        EEG_sensor_information["sampling_frequency"] = payload["sensor_info"]["sampling_frequency"]

   #print(payload["output"])             
    data = payload["output"]['cognitive_analysis']['soft_decision']
   
    # print("soft decision: ", data)
    last_time_message_sent = time.time() - 20
    focus_threshold = 1.8
    distraction_count_threshold = 10
    max_message_rate = 20
    
    
    if data < focus_threshold: 
        print("defocus recognized")
        if distraction_count==0:
            first_disctraction_time = time.time()
        distraction_count +=1
    
    print("distraction count: ", distraction_count)
    
    # if enough distraction events recognized within last minute, and it's been enough time since last notification message, send a notificatino message:
    if distraction_count >= distraction_count_threshold:
        print("distraction count threshold reached!")
        if time.time() - last_time_message_sent > max_message_rate:
            last_time_message_sent = time.time()
            distraction_count = 0
            send_whatsapp("please pay attention!")
            print("please pay attention!")

    
        
    # if not nough distraction event recognized during last munute, reset the disctraction counter:
    if time.time() - first_disctraction_time > 60:
        distraction_count = 0
        
    
    

def EEG_processing_handler2(payload):
    # your implementation
    # print('EEG_processing_handler2')
    pass
    
def customer_gamepad_msg_handler_function(payload):
    # your implementation
    print('customer_gamepad_msg_handler_function')
    pass

def customer_user_data_msg_handler_function(payload):
    # your implementation
    print('customer_user_data_msg_handler_function')
    pass

generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler1, EEG_processing_handler2],
    'gamepad': [customer_gamepad_msg_handler_function],
    'user_data': [customer_user_data_msg_handler_function]
}



def user_data_external_handler(payload):
    username = user_auth.get_username()
    stream_id = payload["stream_id"]
    device_type = payload["device_type"]
    hia_id = payload["hia_id"]
    sensor_info = payload["sensor_info"]
    # print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, hia_id, stream_id, device_type))
    
    # execute each function handler for relevant device_type
    device_type = device_type.lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)
    pass;



# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    global is_tab_open
    global user_auth

    user1_config_path = os.path.join(str(Path().absolute()) ,"config\\","hia_config1.json")
    print(user1_config_path)
    config_user1 = UtilService.load_config_file(user1_config_path)


    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    userRoom.set_device_events_external_handler(user_device_event_external_handler)
    userRoom.connect()
    


    time.sleep(5)
    print(EEG_sensor_information)
    
    
    # time.sleep(60)
    # print("closing tab")
    # whatsapp.close_tab()
    # is_tab_open = False
    
  
if __name__ == '__main__':    
    main()  


# -*- coding: utf-8 -*-
"""
@author: NeuroBrave
this code  is distributed under BSD licence.

this code requires: 
    active internet conenction 
    subscription to NeuroSpeed.io service.
    active EEG stream from neuroSpeed cloud.
"""


GRAPH_ENGAGEMENT = False

from pathlib import Path
import os

from neurospeed.auth.auth_as_user_handler import Auth_AS_User_Handler
from neurospeed.api_socket_handlers.user_room_as_user_handler import UserRoom_AS_User_Handler
from neurospeed.utils.helper_service import UtilService

global user_auth



import statistics
import queue
import time
from vispy import gloo, app, visuals
import numpy as np
import math
from seaborn import color_palette
from scipy.signal import lfilter, lfilter_zi
from mne.filter import create_filter

import socket
import json

stream_q_EEG = queue.Queue(-1)


VERT_SHADER = """
#version 120
// y coordinate of the position.
attribute float a_position;
// row, col, and time index.
attribute vec3 a_index;
varying vec3 v_index;
// 2D scaling factor (zooming).
uniform vec2 u_scale;
// Size of the table.
uniform vec2 u_size;
// Number of samples per signal.
uniform float u_n;
// Color.
attribute vec3 a_color;
varying vec4 v_color;
// Varying variables used for clipping in the fragment shader.
varying vec2 v_position;
varying vec4 v_ab;
void main() {
    float n_rows = u_size.x;
    float n_cols = u_size.y;
    // Compute the x coordinate from the time index.
    float x = -1 + 2*a_index.z / (u_n-1);
    vec2 position = vec2(x - (1 - 1 / u_scale.x), a_position);
    // Find the affine transformation for the subplots.
    vec2 a = vec2(1./n_cols, 1./n_rows)*.9;
    vec2 b = vec2(-1 + 2*(a_index.x+.5) / n_cols,
                    -1 + 2*(a_index.y+.5) / n_rows);
    // Apply the static subplot transformation + scaling.
    gl_Position = vec4(a*u_scale*position+b, 0.0, 1.0);
    v_color = vec4(a_color, 1.);
    v_index = a_index;
    // For clipping test in the fragment shader.
    v_position = gl_Position.xy;
    v_ab = vec4(a, b);
}
"""

FRAG_SHADER = """
#version 120
varying vec4 v_color;
varying vec3 v_index;
varying vec2 v_position;
varying vec4 v_ab;
void main() {
    gl_FragColor = v_color;
    // Discard the fragments between the signals (emulate glMultiDrawArrays).
    if ((fract(v_index.x) > 0.) || (fract(v_index.y) > 0.))
        discard;
    // Clipping test.
    vec2 test = abs((v_position.xy-v_ab.zw)/v_ab.xy);
    if ((test.x > 1))
        discard;
}
"""


global UPDATE_TIMER
UPDATE_TIMER = 1.5

window_title = "use mouse wheel for vertical zoom, +/- keys for horizontal zoom. D key toggles high pass filter"

class Canvas(app.Canvas):
    def __init__(self, stream_info, scale=1000, filt=True):
        app.Canvas.__init__(self, title=window_title, keys='interactive')   
        window = 10
        
        self.sfreq = stream_info['srate']
        self.n_chans = stream_info['n_chans']
        ch_names = stream_info['ch_names']
        
        
        n_samples = int(self.sfreq * window)


        # Number of cols and rows in the table.
        n_rows = self.n_chans
        n_cols = 1

        # Number of signals.
        m = n_rows * n_cols

        # Number of samples per signal.
        n = n_samples

        # Various signal amplitudes.
        amplitudes = np.zeros((m, n)).astype(np.float32)
        # gamma = np.ones((m, n)).astype(np.float32)
        # Generate the signals as a (m, n) array.
        y = amplitudes

        color = color_palette("RdBu_r", n_rows)

        color = np.repeat(color, n, axis=0).astype(np.float32)
        # Signal 2D index of each vertex (row and col) and x-index (sample index
        # within each signal).
        index = np.c_[np.repeat(np.repeat(np.arange(n_cols), n_rows), n),
                      np.repeat(np.tile(np.arange(n_rows), n_cols), n),
                      np.tile(np.arange(n), m)].astype(np.float32)

        self.program = gloo.Program(VERT_SHADER, FRAG_SHADER)
        self.program['a_position'] = y.reshape(-1, 1)
        self.program['a_color'] = color
        self.program['a_index'] = index
        self.program['u_scale'] = (1., 1.)
        self.program['u_size'] = (n_rows, n_cols)
        self.program['u_n'] = n

        # text
        self.font_size = 24.
        self.names = []
        self.quality = []
        for ii in range(self.n_chans):
            text = visuals.TextVisual(ch_names[ii], bold=True, color='white')
            self.names.append(text)
            text = visuals.TextVisual('', bold=True, color='white')
            self.quality.append(text)

        self.quality_colors = color_palette("RdYlGn", 11)[::-1]

        self.scale = scale
        self.n_samples = n_samples
        self.filt = filt
        self.af = [1.0]

        self.data_f = np.zeros((n_samples, self.n_chans))
        self.data = np.zeros((n_samples, self.n_chans))

        self.bf = create_filter(self.data_f.T, self.sfreq, None, 10.,
                                method='fir')

        zi = lfilter_zi(self.bf, self.af)
        self.filt_state = np.tile(zi, (self.n_chans, 1)).transpose()

        self._timer = app.Timer(UPDATE_TIMER, connect=self.on_timer, start=True)
        
        gloo.set_viewport(0, 0, *self.physical_size)
        gloo.set_state(clear_color='black', blend=True,
                       blend_func=('src_alpha', 'one_minus_src_alpha'))
        self.show()

    def on_key_press(self, event):

        # toggle filtering
        if event.key.name == 'D':
            self.filt = not self.filt

        # increase time scale
        if event.key.name in ['+', '-']:
            if event.key.name == '+':
                dx = -0.05
            else:
                dx = 0.05
            scale_x, scale_y = self.program['u_scale']
            scale_x_new, scale_y_new = (scale_x * math.exp(1.0 * dx),
                                        scale_y * math.exp(0.0 * dx))
            self.program['u_scale'] = (
                max(1, scale_x_new), max(1, scale_y_new))
            self.update()

    def on_mouse_wheel(self, event):
        dx = np.sign(event.delta[1]) * .05
        scale_x, scale_y = self.program['u_scale']
        scale_x_new, scale_y_new = (scale_x * math.exp(0.0 * dx),
                                    scale_y * math.exp(2.0 * dx))
        self.program['u_scale'] = (max(1, scale_x_new), max(0.01, scale_y_new))
        self.update()

    def on_timer(self, event):
        """Add some data at the end of each signal (real-time signals)."""       
        samples = stream_q_EEG.get(timeout = 1)
            # samples = [[-1 for j in range(self.n_chans)  ] for i in range(int(self.sfreq * UPDATE_TIMER))]      
        
        if True:
            samples = np.array(samples)[:, ::-1]

            self.data = np.vstack([self.data, samples])
            self.data = self.data[-self.n_samples:]
            filt_samples, self.filt_state = lfilter(self.bf, self.af, samples,
                                                    axis=0, zi=self.filt_state)
            self.data_f = np.vstack([self.data_f, filt_samples])
            self.data_f = self.data_f[-self.n_samples:]

            if self.filt:
                plot_data = self.data_f / self.scale
            elif not self.filt:
                plot_data = (self.data - self.data.mean(axis=0)) / self.scale

            sd = np.std(plot_data[-int(self.sfreq):],
                        axis=0)[::-1] * self.scale
            co = np.int32(np.tanh((sd - 30) / 15) * 5 + 5)
            for ii in range(self.n_chans):
                self.quality[ii].text = '%.2f' % (sd[ii])
                self.quality[ii].color = self.quality_colors[co[ii]]
                self.quality[ii].font_size = 12 + co[ii]

                self.names[ii].font_size = 12 + co[ii]
                self.names[ii].color = self.quality_colors[co[ii]]

            self.program['a_position'].set_data(
                plot_data.T.ravel().astype(np.float32))
            self.update()

    def on_resize(self, event):
        # Set canvas viewport and reconfigure visual transforms to match.
        vp = (0, 0, self.physical_size[0], self.physical_size[1])
        self.context.set_viewport(*vp)

        for ii, t in enumerate(self.names):
            t.transforms.configure(canvas=self, viewport=vp)
            t.pos = (self.size[0] * 0.025,
                     ((ii + 0.5) / self.n_chans) * self.size[1])

        for ii, t in enumerate(self.quality):
            t.transforms.configure(canvas=self, viewport=vp)
            t.pos = (self.size[0] * 0.975,
                     ((ii + 0.5) / self.n_chans) * self.size[1])

    def on_draw(self, event):
        gloo.clear()
        gloo.set_viewport(0, 0, *self.physical_size)
        self.program.draw('line_strip')
        [t.draw() for t in self.names + self.quality]





global EEG_sensor_information
EEG_sensor_information=None
global delta_wave_history
delta_wave_history = []

def EEG_processing_handler1(payload):
    
    '''
    this function is being called every time a packet arrives, depending on confifguration of the HIA source; usually this is once per second.
    the code extracts the useful sensor information that comes with every packet, but does it just one time to be elegant and not keep overwriting the EEG_sensor_information variable   
    
    this function exracts cognitive effort index based on research:
    
    Gvion, Shahaf (2021): "Real-time monitoring of barriers to patient engagement for imporved rehabilitation: a protocol and representative case reports"
    link to this article: https://doi.org/10.1080/17483107.2021.1929513
    
    thanks to Neurospeed, the implementation of engagement index is few lines of code. 
    
    the data is then oversampled 100x and put on queue for collection by the graphing function. 
    the graphing function is called by a timer, collects the data from queue and plots it graphically.
    
     '''  
     
    global delta_wave_history
    global EEG_sensor_information    
    if EEG_sensor_information == None:
        EEG_sensor_information = {}
        EEG_sensor_information["EEG_channel_names"] = payload["sensor_info"]["channel_map"]
        EEG_sensor_information["EEG_channel_num"] = len(payload["sensor_info"]["channel_map"])    
        EEG_sensor_information["sampling_frequency"] = payload["sensor_info"]["sampling_frequency"]



    # electrode_select is a list of EEG electrodes to be included in the calculation of the cognitive effort index. 
    # if a single electrode is desired, create a list with a single value:
    #  if all available electrodes are to be included, use this:
    # electrode_select = "all"
    electrode_select = [0]
      
    
    
    if "all" in electrode_select:
        delta_wave_power = statistics.mean(payload["output"]['brainwave_power']['delta'])
    else:
        delta_wave_power = statistics.mean([payload["output"]['brainwave_power']['delta'][electrode_index] for electrode_index in electrode_select])
    
    delta_wave_history.append(delta_wave_power)
    if len(delta_wave_history) > 20: 
        del delta_wave_history[0]
    
    cognitive_effort_index = statistics.stdev(delta_wave_history) / statistics.mean(delta_wave_history)   
    
    if len(delta_wave_history)< 20:
        print("filling buffer for calculation, please stand by...")
    else: 
        print("engagement value: ", cognitive_effort_index)
        # if cognitive_effort_index>1.0: 
        #     print("value above 1.0 generally indicates noisy data!")
        
    
    if GRAPH_ENGAGEMENT: 
        values = []
        for i in range(100):
            values.append([cognitive_effort_index])
        stream_q_EEG.put(values, timeout = 1)

  


def EEG_processing_handler2(payload):
    # your implementation
    # print('EEG_processing_handler2')
    pass
    
def customer_gamepad_msg_handler_function(payload):
    # your implementation
    print('customer_gamepad_msg_handler_function')
    pass

def customer_user_data_msg_handler_function(payload):
    # your implementation
    print('customer_user_data_msg_handler_function')
    pass

generic_handler = { # allows multiple handlers for each device
    'eeg': [ EEG_processing_handler1, EEG_processing_handler2],
    'gamepad': [customer_gamepad_msg_handler_function],
    'user_data': [customer_user_data_msg_handler_function]
}



def user_data_external_handler(payload):
    username = user_auth.get_username()
    stream_id = payload["stream_id"]
    device_type = payload["device_type"]
    hia_id = payload["hia_id"]
    sensor_info = payload["sensor_info"]
    # print("data from user: {} hia: {} stream_id: {} device_type: {}".format(username, hia_id, stream_id, device_type))
    
    # execute each function handler for relevant device_type
    device_type = device_type.lower()
    handler_functions = generic_handler[device_type]
    for func in handler_functions:
        func(payload)
    
# see README(4) for payload structure  
def user_device_event_external_handler(payload):
    print('event: ', payload)
    pass;



# Hint: you can set "Verbose_socket_log": "True" in user config to enable more logging
def main():
    global user_auth

    user1_config_path = os.path.join(str(Path().absolute()) ,"config\\","hia_config1.json")
    print(user1_config_path)
    config_user1 = UtilService.load_config_file(user1_config_path)


    # Authenticate as user
    user_auth = Auth_AS_User_Handler(config_user1)
    user_auth.login()

    userRoom = UserRoom_AS_User_Handler(user_auth)
    userRoom.set_data_external_handler(user_data_external_handler)
    userRoom.set_device_events_external_handler(user_device_event_external_handler)

    userRoom.connect()
    



    time.sleep(5)
    print(EEG_sensor_information)
    
    if GRAPH_ENGAGEMENT: 
        stream_info = {'n_chans': 1,'srate': 100  , 'ch_names': ["engamement"]    }
        Canvas(stream_info)
        app.run()



    

if __name__ == '__main__':    
    main()  